#!/usr/bin/env python

# vim: set expandtab shiftwidth=4 tabstop=4:

#
# Jacek Fedorynski <jfedor@jfedor.org>
# http://www.jfedor.org/
#

import serial, time, datetime, re, sys, decimal, math, tempfile, os, webbrowser
import pygame

execfile('config.txt', globals())

s = serial.Serial()
s.port = port
s.baudrate = baudrate
s.open()

gps = {'date': '', 'course': '', 'satellites': [], 'altitude': '', 'longitude': ['', ''], 'used satellites': [], 'time': '', 'latitude': ['', ''], 'speed': '', 'fix': ''}
partialsat = []
color = (255, 0, 0)
vmax = decimal.Decimal("0")
updatecoords = 0
updatefix = 0
tagopen = 0

def start_files():
    global log, track
    now = datetime.datetime.now().isoformat()
    now = re.sub(r':', r'', now)
    log = file(os.path.join('logs', 'log.' + now + '.txt'), 'ab')
    track = file(os.path.join('tracks', 'track.' + now + '.kml'), 'wb')
    track.write("""<?xml version="1.0" ?>
<kml xmlns="http://earth.google.com/kml/2.0">
<Document>
<Style id="track">
  <LineStyle>
    <color>ff0000ff</color>
    <width>3</width>
  </LineStyle>
</Style>
<Placemark>
  <styleUrl>#track</styleUrl>
  <name>""")
    track.write(now)
    track.write("""</name>
  <MultiGeometry>
""")

def stop_files():
    global log, track, tagopen
    log.close()
    if tagopen:
        close_tag()
    track.write("""  </MultiGeometry>
</Placemark>
</Document>
</kml>
""")
    track.close()

def open_tag():
    global track, tagopen
    track.write("""<LineString>
<tessellate>1</tessellate>
<coordinates>
""")
    tagopen = 1

def close_tag():
    global track, tagopen
    track.write("""</coordinates>
</LineString>
""")
    tagopen = 0

def valid_checksum(text, checksum):
    sum = 0
    for i in range(len(text)):
        sum ^= ord(text[i])
    return sum == int(checksum, 16)

def write_kml():
    global gps, track, tagopen
    coords = str(parse_coord(gps['latitude'])) + ',' + str(parse_coord(gps['longitude'])) + ',0'
    if tagopen:
        track.write(coords + "\n")
    fd, name = tempfile.mkstemp(dir='.')
    os.close(fd)
    f = file(name, "wb")
    f.write("""<?xml version="1.0" ?>
<kml xmlns="http://earth.google.com/kml/2.0">
<Document>
<Placemark>
  <name>You are here</name>
  <Point>
    <coordinates>""")
    f.write(coords)
    f.write("""</coordinates>
  </Point>
</Placemark>
</Document>
</kml>
""")
    f.close()
    if os.name != 'posix':
        try:
            os.unlink('pos.kml')
        except:
            pass
    os.rename(name, 'pos.kml')

def process_line(f):
    global gps, partialsat, updatecoords, updatefix
    if f[0] == 'GPGGA':
        gps['time'] = f[1]
        gps['altitude'] = f[9]
        gps['latitude'] = f[4:6]
        gps['longitude'] = f[2:4]
        updatecoords = 1
    elif f[0] == 'GPGSA':
        gps['fix'] = f[2]
        gps['used satellites'] = []
        for i in range(3,14):
            if f[i] != '':
                gps['used satellites'].append(f[i])
        updatefix = 1
    elif f[0] == 'GPRMC':
        gps['speed'] = f[7]
        gps['course'] = f[8]
        gps['date'] = f[9]
    elif f[0] == 'GPGSV':
        if int(f[2]) == 1:
            partialsat = []
        partialsat.append(f[4:8])
        partialsat.append(f[8:12])
        partialsat.append(f[12:16])
        partialsat.append(f[16:20])
        if f[2] == f[1]:
            gps['satellites'] = partialsat[:int(f[3])]

def process_events():
    global screen, vmax, gps
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            quit()
        elif event.type == pygame.KEYDOWN:
            if event.key == pygame.K_ESCAPE:
                quit()
            elif event.key == pygame.K_r:
                vmax = decimal.Decimal("0")
            elif event.key == pygame.K_g:
                url = 'http://maps.google.com/maps?q=' + str(parse_coord(gps['longitude'])) + ',' + str(parse_coord(gps['latitude']))
                webbrowser.open(url)
            elif event.key == pygame.K_h:
                url = 'file://' + os.path.join(os.getcwd(), 'gps.kml')
                webbrowser.open(url)
            elif event.key == pygame.K_n:
                stop_files()
                start_files()

def quit():
    global s
    stop_files()
    s.close()
    sys.exit()

def parse_coord(c):
    m = re.search(r'^(\d{1,3})(\d{2}\.\d+)$', c[0])
    if m:
        tmp = decimal.Decimal(m.group(1)) + decimal.Decimal(m.group(2))/decimal.Decimal("60")
        if c[1] == 'S' or c[1] == 'W':
            tmp *= decimal.Decimal("-1")
        return tmp
    else:
        return decimal.Decimal("0")

def draw_text(text, pos, font = None):
    global screen, font1, color
    if font is None:
        font = font1
    pic = font.render(text, 1, color)
    screen.blit(pic, pos)

def format_date(d2):
    d = '%6s' % d2
    return '20' + d[4:6] + '-' + d[2:4] + '-' + d[0:2]

def format_time(t2):
    t = '%6s' % t2
    return t[0:2] + ':' + t[2:4] + ':' + t[4:6]

def format_coord(coord):
    c = '%10s' % coord[0]
    return c[0:3] + '\xb0' + c[3:] + "'" + coord[1]

def format(x):
    return re.sub(r'^0+(?!\.)', r'', x)

def angle2pos(x, y, r, a):
    return (x-r*math.sin(a), y+r*math.cos(a))

def draw_screen():
    global gps, screen, color, vmax, font1, font2
    screen.fill((0, 0, 0))
    if gps['fix'] == '3':
        color = (0, 255, 0)
    elif gps['fix'] == '2':
        color = (255, 255, 0)
    else:
        color = (255, 0, 0)
    draw_text(format_date(gps['date']), (30, 30))
    draw_text(format_time(gps['time']), (340, 30))
    draw_text(format_coord(gps['longitude']), (30, 70), font2)
    draw_text(format_coord(gps['latitude']), (30, 120), font2)
    draw_text('        HEADING: ' + format(gps['course']) + '\xb0', (30, 340))
    draw_text('       ALTITUDE: ' + format(gps['altitude']) + ' M', (30, 360))
    if gps['speed'] != '':
        speed = decimal.Decimal(gps['speed']) * decimal.Decimal('1.852')
        speedstr = '%.1f' % speed
    else:
        speed = decimal.Decimal('0')
        speedstr = ''
    draw_text('          SPEED: ' + speedstr + ' KM/H', (30, 380))
    if speed > vmax:
        vmax = speed
    draw_text('      TOP SPEED: %.1f KM/H' % vmax, (30, 400))
    if gps['course'] != '':
        h = float(gps['course'])/180.0*math.pi
        pygame.draw.aalines(screen, color, 1, (angle2pos(230, 260, 30, h+0), angle2pos(230, 260, 55, h+30.0/180*math.pi), angle2pos(230, 260, 55, h+math.pi), angle2pos(230, 260, 55, h+330.0/180*math.pi)))
    for i in range(len(gps['satellites'])):
        sat = gps['satellites'][i]
        draw_text(sat[0], (30 + i * 30, 550))
        if sat[0] in gps['used satellites']:
            border = 0
        else:
            border = 1
        if sat[3] != '':
            pygame.draw.rect(screen, color, pygame.Rect(30 + i * 30, 550 - int(sat[3]), 24, int(sat[3])), border)
    pygame.display.flip()

# main

decimal.getcontext().prec = 9
start_files()

pygame.init()
pygame.display.set_caption('Trajeto')
screen = pygame.display.set_mode((460, 600))
font1 = pygame.font.Font("VeraMono.ttf", 20)
font2 = pygame.font.Font("VeraMono.ttf", 48)

while 1:
    line = s.readline()
    log.write(line)
    m = re.search(r'^\$(.*)\*([0-9A-F]{2})\015\012$', line)
    if not m:
        print "Invalid line: " + line
        continue
    if not valid_checksum(m.group(1), m.group(2)):
        print "Invalid checksum: " + line
        continue
    f = m.group(1).split(",")
    process_line(f)
    if updatecoords:
        write_kml()
        updatecoords = 0
    if updatefix:
        if tagopen and not gps['fix'] in ['2', '3']:
            close_tag()
        elif not tagopen and gps['fix'] in ['2', '3']:
            open_tag()
        updatefix = 0
    process_events()
    draw_screen()
